import json
from dataclasses import asdict, dataclass
from functools import reduce
from threading import Condition, Lock, Thread
from typing import Callable, Dict, List
from urllib.parse import urlparse

import websocket

from ._version import __version__
from .config import Config
from .interceptors.types import InterceptorHandler, Interceptors


@dataclass
class RequestObject:
    # Identifier established by the initiator of the request
    id: int
    # name of the engine method.
    method: str
    # target of the method.
    handle: int
    # the parameters can be provided by name through an object or by position through an array
    params: any
    # version of JSON-RPC defaults to 2.0
    jsonrpc: str = "2.0"


@dataclass
class ResponseObject:
    # id of the backend object.
    id: int
    #  QIX type of the backend object. Can for example be "Doc" or "GenericVariable".
    type: str
    # Custom type of the backend object, if defined in qInfo.
    genericType: str
    # Handle of the backend object.
    handle: int
    # represents the returned value from engine
    result: dict


RequestInterceptor = Callable[[RequestObject], RequestObject]
""" RPC Request interceptor """
ResponseInterceptor = Callable[[dict], dict]
""" RPC Request interceptor """


class FailedToConnect(Exception):
    pass


class RpcSession:

    _interceptors: Interceptors[ResponseInterceptor, RequestInterceptor]

    _ws_url: str
    _headers: List[str]

    _socket = None
    _watch_recv_thread = None

    def __init__(
        self, ws_url: str, headers: List[str] = None, interceptors: Interceptors = None
    ):
        if headers is None:
            headers = []
        if not ws_url:
            raise Exception("Empty url")
        self._headers = headers
        self._ws_url = ws_url
        self._interceptors = interceptors
        self.lock = Lock()

    def _watch_recv(self):
        """
        _watch_recv watches for socket responses.
        Adds the response to _received.
        """

        while True:
            if not self.is_connected():
                return
            try:
                res = self._socket.recv()
            except Exception as err:
                self._socket = None
                res = False
                self._recv_error = err
            with self._received_added:
                if res:
                    res = json.loads(res)
                    # add response to _received and notify waiting
                    if "id" in res:
                        self._received[res["id"]] = res
                        self._received_added.notify_all()
                else:
                    # notify waiting receivers so that
                    # the not connected error can be raised
                    # if the error is raised from here then the
                    # wait_response will never finish
                    self._received_added.notify_all()

    def open(self):
        """
        connect establishes a connection to provided url
        using the specified headers.

        If the client is already connected an exception will
        be raised.
        """
        if self.is_connected():
            raise Exception("Client already connected")
        socket = websocket.WebSocket()
        try:
            socket.connect(self._ws_url, header=self._headers, suppress_origin=True)
        except Exception as exc:
            raise FailedToConnect() from exc

        self._socket = socket
        self._received = {}
        self._id = -1
        self._received_added = Condition()

        self._watch_recv_thread = Thread(target=self._watch_recv)
        self._watch_recv_thread.start()
        return self

    def is_connected(self):
        """
        return connected state
        """
        return self._socket and self._socket.connected

    def close(self):
        """
        close closes the socket (if it's open).
        """

        if self.is_connected():
            self._socket.send_close()
        if self._watch_recv_thread is not None and self._watch_recv_thread.is_alive():
            self._watch_recv_thread.join()

    def __enter__(self):
        """
        __enter__ is called when client is used in a 'with' statement.
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        __exit__ is called when the 'with' scope is exited. This will call
        the client's close method.
        """

        self.close()

    def send(self, method: str, handle: int, *args, **kwargs):
        """
        send is a thread-safe method that sends a websocket-message with the
        specified method, handle and parameters.
        The resulting response is returned.

        If the client isn't connected an exception is raised.

        Parameters
        ----------
        method: str
            engine method name for the request
        handle: int
            the associated handle
        args: tuple
            parameters by position in an array
        kwargs: dict
            parameters provided by name in an object
        """

        if not self.is_connected():
            raise Exception("Client not connected")

        self.lock.acquire()
        self._id += 1
        id_ = self._id
        self.lock.release()

        data = RequestObject(
            id=self._id,
            method=method,
            handle=handle,
            params=args if not kwargs else kwargs,
        )

        # send and wait respons
        data = reduce(lambda d, f: f(d), self._interceptors["request"].handlers, data)
        json_data = json.dumps(asdict(data))
        self._socket.send(json_data)
        res = self._wait_response(id_)
        res["request_data"] = data
        res = reduce(lambda r, f: f(r), self._interceptors["response"].handlers, res)
        return_value = None
        if "result" in res:
            return_value = res["result"]
        elif "error" in res:
            raise Exception(res["error"]["message"])
        else:
            return_value = res
        return return_value

    def _wait_response(self, id_):
        """
        _wait_response waits (blocking) for a message with the specified id.
        Internal method that should only be called from send
        """

        with self._received_added:
            while id_ not in self._received:
                if not self.is_connected():
                    if self._recv_error:
                        raise self._recv_error
                    else:
                        raise Exception("not connected")
                self._received_added.wait()
            res = self._received[id_]
            del self._received[id_]
            return res


class RpcClient:
    __config: Config
    # property for storing the interceptors
    interceptors: Interceptors[ResponseInterceptor, RequestInterceptor]

    sessions: Dict[str, RpcSession] = {}

    def __init__(self, config) -> None:
        self.__config = config
        # initiating the interceptors
        self.interceptors = dict(
            request=InterceptorHandler[RequestInterceptor](),
            response=InterceptorHandler[ResponseInterceptor](),
        )

    def rpc(self, app_id: str) -> RpcSession:
        hostname = urlparse(self.__config.host).hostname
        ws_url = "wss://" + hostname.strip("/") + "/app/" + app_id
        version = __version__[1:]
        headers = [
            "Authorization: Bearer %s" % self.__config.api_key,
            f"User-Agent: qlik-sdk-python/{version}",
        ]
        if ws_url not in self.sessions:
            self.sessions[ws_url] = RpcSession(ws_url, headers, self.interceptors)
        return self.sessions[ws_url]


class RpcClientInstance:
    interceptors: Interceptors

    def __init__(self, rpcClient: RpcClient) -> None:
        self._rpcClient = rpcClient
        self.interceptors = rpcClient.interceptors

    def __call__(self, app_id: str) -> RpcSession:
        return self._rpcClient.rpc(app_id)



import typing

import requests

from ._version import __version__
from .config import Config
from .rest import RestClient, RestClientInstance
from .rpc import RpcClient, RpcClientInstance


class Auth:
    """
    Auth can be used to make rest and rpc calls

    Parameters
    ----------
    config: Config
        the required configuration object

    Examples
    --------
    >>> from qlik_sdk import Auth, AuthType, Config
    ...
    ... clients = Auth(Config(host=base_url, auth_type=AuthType.APIKey, api_key=api_key))
    ... get_users_res = clients.rest(path="/users/me")
    """

    config: Config
    rest: RestClientInstance
    """
    rest method can be used to make raw calls against Qlik Cloud

    Parameters
    ----------
    app_id: str
    method: str, default GET
        string HTTP verb
    path: str
        representing the api endpoint ex: `/users/me`
    data: dict, optional
        Dictionary, list of tuples, bytes, or file-like object to send in the body of the Request.
    params: dict, optional
        Dictionary, list of tuples or bytes to send in the query string for the Request.
    headers: dict, optional
        Dictionary of HTTP Headers to send with the Request
    files: dict, optional
        Dictionary of {filename: fileobject} files to multipart upload.

    Attributes
    ----------
    interceptors: Interceptors

    Examples
    ----------
    >>> auth = Auth(Config(host=self.base_url, auth_type=AuthType.APIKey, api_key=self.api_key))
    ... user_me = auth.rest(path="/users/me")
    ...
    # And with interceptors.
    >>> auth = Auth(Config(host=self.base_url, auth_type=AuthType.APIKey, api_key=self.api_key))
    ... def log_req(req: requests.Request) -> requests.Request:
    ...     print(req)
    ...     return req
    ...
    ... auth.rpc.interceptors["request"].use(log_req)
    ... app_list = auth.rest(path="/items", params={"resourceType":"app", "limit": 100})
    """

    rpc: RpcClientInstance
    """
    rpc returns an RpcClient that can be used to
    connect to the engine for a specific app

    Parameters
    ----------
    app_id: str

    Attributes
    ----------
    interceptors: Interceptors

    Examples
    ----------
    >>> rpc_session = auth.rpc(app_id=session_app_id)
    ... with rpc_session.opn() as rpc_client:
    ...     app = rpc_client.send("OpenDoc", -1, session_app_id)
    ...
    # And with interceptors.
    >>> auth.rpc.interceptors["request"].use(log_req)
    ... rpc_session = auth.rpc(app_id=session_app_id)
    ...
    ... with rpc_session.open() as rpc_client:
    ...     app = rpc_client.send("OpenDoc", -1, session_app_id)
    """
    paginate_rest: typing.Callable[[str, str, typing.Dict, any], typing.Iterator[any]]

    def __init__(self, config: Config):
        config.validate()
        self.config = config
        self.rest = RestClientInstance(RestClient(config))

        # register the api-key-auth interceptor
        def add_api_key(req: requests.Request) -> requests.Request:
            req.headers["authorization"] = "Bearer " + self.config.api_key
            return req

        # register the user-agent interceptor
        def add_user_agent(req: requests.Request):
            # version without v in user-agent
            # consistent with format of version in qlik-cli-user-agent
            version = __version__[1:]
            req.headers["User-Agent"] = "qlik-sdk-python/" + version
            return req

        self.rest.interceptors["request"].use([add_api_key, add_user_agent])
        self.rpc_client = RpcClient(self.config)
        self.rpc = RpcClientInstance(self.rpc_client)
        self.paginate_rest = self.rest.paginate_rest



# This is spectacularly generated code by spectacular v0.0.0 based on
# Qlik Cloud Services 0.348.2

from __future__ import annotations

import io
import json
from dataclasses import asdict, dataclass

from ..auth import Auth, Config
from ..listable import ListableResource


@dataclass
class DataFileUploadResponse:
    """

    Attributes
    ----------
    createdDate: str
      The date that the uploaded file was created.
    id: str
      The ID for the uploaded file.
    name: str
      The name of the uploaded file.
    ownerId: str
      The 'owner' of a file is the user who last uploaded the file's content.
    size: int
      The size of the uploaded file, in bytes.
    appId: str
      If this file is bound to the lifecycle of a specific app, this is the ID of this app.
    modifiedDate: str
      The date that the updated file was last modified.
    spaceId: str
      If the file was uploaded to a team space, this is the ID of that space.
    """

    createdDate: str = None
    id: str = None
    name: str = None
    ownerId: str = None
    size: int = None
    appId: str = None
    modifiedDate: str = None
    spaceId: str = None

    def __init__(self_, **kvargs):
        if "createdDate" in kvargs:
            if (
                type(kvargs["createdDate"]).__name__
                is self_.__annotations__["createdDate"]
            ):
                self_.createdDate = kvargs["createdDate"]
            else:
                self_.createdDate = kvargs["createdDate"]
        if "id" in kvargs:
            if type(kvargs["id"]).__name__ is self_.__annotations__["id"]:
                self_.id = kvargs["id"]
            else:
                self_.id = kvargs["id"]
        if "name" in kvargs:
            if type(kvargs["name"]).__name__ is self_.__annotations__["name"]:
                self_.name = kvargs["name"]
            else:
                self_.name = kvargs["name"]
        if "ownerId" in kvargs:
            if type(kvargs["ownerId"]).__name__ is self_.__annotations__["ownerId"]:
                self_.ownerId = kvargs["ownerId"]
            else:
                self_.ownerId = kvargs["ownerId"]
        if "size" in kvargs:
            if type(kvargs["size"]).__name__ is self_.__annotations__["size"]:
                self_.size = kvargs["size"]
            else:
                self_.size = kvargs["size"]
        if "appId" in kvargs:
            if type(kvargs["appId"]).__name__ is self_.__annotations__["appId"]:
                self_.appId = kvargs["appId"]
            else:
                self_.appId = kvargs["appId"]
        if "modifiedDate" in kvargs:
            if (
                type(kvargs["modifiedDate"]).__name__
                is self_.__annotations__["modifiedDate"]
            ):
                self_.modifiedDate = kvargs["modifiedDate"]
            else:
                self_.modifiedDate = kvargs["modifiedDate"]
        if "spaceId" in kvargs:
            if type(kvargs["spaceId"]).__name__ is self_.__annotations__["spaceId"]:
                self_.spaceId = kvargs["spaceId"]
            else:
                self_.spaceId = kvargs["spaceId"]
        for k, v in kvargs.items():
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)

    def get_connection(self) -> ConnectionsResponse:
        """
        Get the built-in connection used by the engine to load/write data files given a connection ID.

        Parameters
        ----------
        """

        response = self.auth.rest(
            path="/data-files/connections/{id}".replace("{id}", self.id),
            method="GET",
            params={},
            data=None,
        )
        obj = ConnectionsResponse(**response.json())
        obj.auth = self.auth
        return obj

    def delete(self) -> None:
        """
        Delete the specified data file.

        Parameters
        ----------
        """

        self.auth.rest(
            path="/data-files/{id}".replace("{id}", self.id),
            method="DELETE",
            params={},
            data=None,
        )

    def set(
        self,
        File: io.BufferedReader,
        appId: str = None,
        connectionId: str = None,
        name: str = None,
        sourceId: str = None,
        tempContentFileId: str = None,
    ) -> DataFileUploadResponse:
        """
        Re-upload an existing data file.

        Parameters
        ----------
        """

        files_dict = {}
        files_dict["File"] = File

        Json_dict = {}
        if appId is not None:
            Json_dict["appId"] = appId
        if connectionId is not None:
            Json_dict["connectionId"] = connectionId
        if name is not None:
            Json_dict["name"] = name
        if sourceId is not None:
            Json_dict["sourceId"] = sourceId
        if tempContentFileId is not None:
            Json_dict["tempContentFileId"] = tempContentFileId
        files_dict["Json"] = (None, json.dumps(Json_dict))

        response = self.auth.rest(
            path="/data-files/{id}".replace("{id}", self.id),
            method="PUT",
            params={},
            data=None,
            files=files_dict,
        )
        self.__init__(**response.json())
        return self

    def change_owner(self, data: ChangeDataFileOwnerRequest) -> None:
        """
        Change the owner of an existing data file.
        This is primarily an admin type of operation. In general, the owner of a data file is implicitly set as
        part of a data file upload. For data files that reside in a personal space, changing the owner has the
        effect of moving the data file to the new owner's personal space.

        Parameters
        ----------
        data: ChangeDataFileOwnerRequest
        """

        try:
            data = asdict(data)
        except:
            data = data

        self.auth.rest(
            path="/data-files/{id}/actions/change-owner".replace("{id}", self.id),
            method="POST",
            params={},
            data=data,
        )

    def change_space(self, data: ChangeDataFileSpaceRequest) -> None:
        """
        Change the space that an existing data file resides in.
        This is to allow for a separate admin type of operation that is more global in terms of access in cases
        where admin users may not explicitly have been granted full access to a given space within the declared
        space-level permissions. If the space ID is set to null, then the datafile will end up residing in the
        personal space of the user who is the owner of the file.

        Parameters
        ----------
        data: ChangeDataFileSpaceRequest
        """

        try:
            data = asdict(data)
        except:
            data = data

        self.auth.rest(
            path="/data-files/{id}/actions/change-space".replace("{id}", self.id),
            method="POST",
            params={},
            data=data,
        )


@dataclass
class BatchChangeSpaceItem:
    """

    Attributes
    ----------
    id: str
      The ID of the data file whose space will be changed.
    spaceId: str
      The ID of the new space. Passing in a null will result in the data file being moved to the user's
      personal space.
    """

    id: str = None
    spaceId: str = None

    def __init__(self_, **kvargs):
        if "id" in kvargs:
            if type(kvargs["id"]).__name__ is self_.__annotations__["id"]:
                self_.id = kvargs["id"]
            else:
                self_.id = kvargs["id"]
        if "spaceId" in kvargs:
            if type(kvargs["spaceId"]).__name__ is self_.__annotations__["spaceId"]:
                self_.spaceId = kvargs["spaceId"]
            else:
                self_.spaceId = kvargs["spaceId"]
        for k, v in kvargs.items():
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class BatchDeleteItem:
    """

    Attributes
    ----------
    id: str
      The ID of the data file to delete.
    """

    id: str = None

    def __init__(self_, **kvargs):
        if "id" in kvargs:
            if type(kvargs["id"]).__name__ is self_.__annotations__["id"]:
                self_.id = kvargs["id"]
            else:
                self_.id = kvargs["id"]
        for k, v in kvargs.items():
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ChangeDataFileOwnerRequest:
    """

    Attributes
    ----------
    ownerId: str
      The ID of the new owner.
    """

    ownerId: str = None

    def __init__(self_, **kvargs):
        if "ownerId" in kvargs:
            if type(kvargs["ownerId"]).__name__ is self_.__annotations__["ownerId"]:
                self_.ownerId = kvargs["ownerId"]
            else:
                self_.ownerId = kvargs["ownerId"]
        for k, v in kvargs.items():
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ChangeDataFileSpaceRequest:
    """

    Attributes
    ----------
    spaceId: str
      The ID of the space. If null, this data file will be moved to the user's personal space.
    """

    spaceId: str = None

    def __init__(self_, **kvargs):
        if "spaceId" in kvargs:
            if type(kvargs["spaceId"]).__name__ is self_.__annotations__["spaceId"]:
                self_.spaceId = kvargs["spaceId"]
            else:
                self_.spaceId = kvargs["spaceId"]
        for k, v in kvargs.items():
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ConnectionsResponse:
    """

    Attributes
    ----------
    connectStatement: str
      The connect statement that will be passed to the connector when invoked.
    id: str
    name: str
      The name of the connection.
    type: str
      The type of the connection.
    spaceId: str
      The team space that the given connection is associated with. If null, the connection is not associated
      with any specific team space.
    """

    connectStatement: str = None
    id: str = None
    name: str = None
    type: str = None
    spaceId: str = None

    def __init__(self_, **kvargs):
        if "connectStatement" in kvargs:
            if (
                type(kvargs["connectStatement"]).__name__
                is self_.__annotations__["connectStatement"]
            ):
                self_.connectStatement = kvargs["connectStatement"]
            else:
                self_.connectStatement = kvargs["connectStatement"]
        if "id" in kvargs:
            if type(kvargs["id"]).__name__ is self_.__annotations__["id"]:
                self_.id = kvargs["id"]
            else:
                self_.id = kvargs["id"]
        if "name" in kvargs:
            if type(kvargs["name"]).__name__ is self_.__annotations__["name"]:
                self_.name = kvargs["name"]
            else:
                self_.name = kvargs["name"]
        if "type" in kvargs:
            if type(kvargs["type"]).__name__ is self_.__annotations__["type"]:
                self_.type = kvargs["type"]
            else:
                self_.type = kvargs["type"]
        if "spaceId" in kvargs:
            if type(kvargs["spaceId"]).__name__ is self_.__annotations__["spaceId"]:
                self_.spaceId = kvargs["spaceId"]
            else:
                self_.spaceId = kvargs["spaceId"]
        for k, v in kvargs.items():
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class DataFileBatchChangeSpaceRequest:
    """
    Specifies the list of data file change space operations in a single batch.

    Attributes
    ----------
    change-space: list[BatchChangeSpaceItem]
      The list of data files to delete.
    """

    change_space: list[BatchChangeSpaceItem] = None

    def __init__(self_, **kvargs):
        if "change-space" in kvargs:
            if (
                type(kvargs["change-space"]).__name__
                is self_.__annotations__["change-space"]
            ):
                self_.change_space = kvargs["change-space"]
            else:
                self_.change_space = [
                    BatchChangeSpaceItem(**e) for e in kvargs["change-space"]
                ]
        for k, v in kvargs.items():
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class DataFileBatchDeleteRequest:
    """
    Specifies the list of data files to be deleted in a single batch.

    Attributes
    ----------
    delete: list[BatchDeleteItem]
      The list of data files to delete.
    """

    delete: list[BatchDeleteItem] = None

    def __init__(self_, **kvargs):
        if "delete" in kvargs:
            if type(kvargs["delete"]).__name__ is self_.__annotations__["delete"]:
                self_.delete = kvargs["delete"]
            else:
                self_.delete = [BatchDeleteItem(**e) for e in kvargs["delete"]]
        for k, v in kvargs.items():
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class GetConnectionsResponse:
    """

    Attributes
    ----------
    data: list[ConnectionsResponse]
    links: LinksResponse
    """

    data: list[ConnectionsResponse] = None
    links: LinksResponse = None

    def __init__(self_, **kvargs):
        if "data" in kvargs:
            if type(kvargs["data"]).__name__ is self_.__annotations__["data"]:
                self_.data = kvargs["data"]
            else:
                self_.data = [ConnectionsResponse(**e) for e in kvargs["data"]]
        if "links" in kvargs:
            if type(kvargs["links"]).__name__ is self_.__annotations__["links"]:
                self_.links = kvargs["links"]
            else:
                self_.links = LinksResponse(**kvargs["links"])
        for k, v in kvargs.items():
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class GetDataFileInfosResponse:
    """

    Attributes
    ----------
    data: list[DataFileUploadResponse]
    links: LinksResponse
    """

    data: list[DataFileUploadResponse] = None
    links: LinksResponse = None

    def __init__(self_, **kvargs):
        if "data" in kvargs:
            if type(kvargs["data"]).__name__ is self_.__annotations__["data"]:
                self_.data = kvargs["data"]
            else:
                self_.data = [DataFileUploadResponse(**e) for e in kvargs["data"]]
        if "links" in kvargs:
            if type(kvargs["links"]).__name__ is self_.__annotations__["links"]:
                self_.links = kvargs["links"]
            else:
                self_.links = LinksResponse(**kvargs["links"])
        for k, v in kvargs.items():
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class LinkResponse:
    """

    Attributes
    ----------
    href: str
    """

    href: str = None

    def __init__(self_, **kvargs):
        if "href" in kvargs:
            if type(kvargs["href"]).__name__ is self_.__annotations__["href"]:
                self_.href = kvargs["href"]
            else:
                self_.href = kvargs["href"]
        for k, v in kvargs.items():
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class LinksResponse:
    """

    Attributes
    ----------
    next: LinkResponse
    prev: LinkResponse
    self: LinkResponse
    """

    next: LinkResponse = None
    prev: LinkResponse = None
    self: LinkResponse = None

    def __init__(self_, **kvargs):
        if "next" in kvargs:
            if type(kvargs["next"]).__name__ is self_.__annotations__["next"]:
                self_.next = kvargs["next"]
            else:
                self_.next = LinkResponse(**kvargs["next"])
        if "prev" in kvargs:
            if type(kvargs["prev"]).__name__ is self_.__annotations__["prev"]:
                self_.prev = kvargs["prev"]
            else:
                self_.prev = LinkResponse(**kvargs["prev"])
        if "self" in kvargs:
            if type(kvargs["self"]).__name__ is self_.__annotations__["self"]:
                self_.self = kvargs["self"]
            else:
                self_.self = LinkResponse(**kvargs["self"])
        for k, v in kvargs.items():
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class QuotaResponse:
    """

    Attributes
    ----------
    allowedExtensions: list[str]
      The allowed file extensions on files that are uploaded.
    allowedInternalExtensions: list[str]
      The allowed file extensions for files that are only used internally by the system (and thus not typically
      shown to end users).
    maxFileSize: int
      Maximum allowable size of an uploaded file.
    maxLargeFileSize: int
      Maximum allowable size for a single uploaded large data file (in bytes). This is a file that was indirectly
      uploaded using the temp content service chunked upload capability.
    maxSize: int
      The maximum aggregate size of all files uploaded by a given user.
    size: int
      The current aggregate size of all files uploaded by a given user. If the current aggregate size is greater
      than the maximum aggregate size, this is a quota violation.
    """

    allowedExtensions: list[str] = None
    allowedInternalExtensions: list[str] = None
    maxFileSize: int = None
    maxLargeFileSize: int = None
    maxSize: int = None
    size: int = None

    def __init__(self_, **kvargs):
        if "allowedExtensions" in kvargs:
            if (
                type(kvargs["allowedExtensions"]).__name__
                is self_.__annotations__["allowedExtensions"]
            ):
                self_.allowedExtensions = kvargs["allowedExtensions"]
            else:
                self_.allowedExtensions = kvargs["allowedExtensions"]
        if "allowedInternalExtensions" in kvargs:
            if (
                type(kvargs["allowedInternalExtensions"]).__name__
                is self_.__annotations__["allowedInternalExtensions"]
            ):
                self_.allowedInternalExtensions = kvargs["allowedInternalExtensions"]
            else:
                self_.allowedInternalExtensions = kvargs["allowedInternalExtensions"]
        if "maxFileSize" in kvargs:
            if (
                type(kvargs["maxFileSize"]).__name__
                is self_.__annotations__["maxFileSize"]
            ):
                self_.maxFileSize = kvargs["maxFileSize"]
            else:
                self_.maxFileSize = kvargs["maxFileSize"]
        if "maxLargeFileSize" in kvargs:
            if (
                type(kvargs["maxLargeFileSize"]).__name__
                is self_.__annotations__["maxLargeFileSize"]
            ):
                self_.maxLargeFileSize = kvargs["maxLargeFileSize"]
            else:
                self_.maxLargeFileSize = kvargs["maxLargeFileSize"]
        if "maxSize" in kvargs:
            if type(kvargs["maxSize"]).__name__ is self_.__annotations__["maxSize"]:
                self_.maxSize = kvargs["maxSize"]
            else:
                self_.maxSize = kvargs["maxSize"]
        if "size" in kvargs:
            if type(kvargs["size"]).__name__ is self_.__annotations__["size"]:
                self_.size = kvargs["size"]
            else:
                self_.size = kvargs["size"]
        for k, v in kvargs.items():
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


class DataFiles:
    def __init__(self, config: Config) -> None:
        self.config = config
        self.auth = Auth(config)

    def get_data_files(
        self,
        connectionId: str = None,
        appId: str = None,
        name: str = None,
        page: str = None,
        limit: int = 20,
        ownerId: str = None,
        allowInternalFiles: bool = None,
        sort: str = None,
        max_items: int = 20,
    ) -> ListableResource[DataFileUploadResponse]:
        """
        Get descriptive info for the specified data files.


        connectionId: str
          Return files that reside in the space referenced by the specified DataFiles connection. If this parameter
          is not specified, the user's personal space is implied.

        appId: str
          Only return files scoped to the specified app. If this parameter is not specified, only files that are not
          scoped to any app are returned. "*" implies all app-scoped files are returned.

        name: str
          Filter the list of files returned to the given file name.

        page: str
          If present, the cursor that starts the page of data that is returned.

        limit: int
          If present, the maximum number of data files to return.

        ownerId: str
          If present, fetch the data files for the specified owner. If a connectionId is specified in this case, the
          returned list is constrained to the specified space. If connectionId is not specified, then all files owned
          by the specified user are returned regardless of the personal space that a given file resides in.

        allowInternalFiles: bool
          If set to false, do not return data files with internal extensions else return all the data files.

        sort: str
          The name of the field used to sort the result. By default, the sort order is ascending. Putting a '+' prefix on
          the sort field name explicitly indicates ascending sort order. A '-' prefix indicates a descending sort order.

        Parameters
        ----------
        connectionId: str = None
        appId: str = None
        name: str = None
        page: str = None
        limit: int = 20
        ownerId: str = None
        allowInternalFiles: bool = None
        sort: str = None
        """
        query_params = {}
        if connectionId is not None:
            query_params["connectionId"] = connectionId
        if appId is not None:
            query_params["appId"] = appId
        if name is not None:
            query_params["name"] = name
        if page is not None:
            query_params["page"] = page
        if limit is not None:
            query_params["limit"] = limit
        if ownerId is not None:
            query_params["ownerId"] = ownerId
        if allowInternalFiles is not None:
            query_params["allowInternalFiles"] = allowInternalFiles
        if sort is not None:
            query_params["sort"] = sort

        response = self.auth.rest(
            path="/data-files",
            method="GET",
            params=query_params,
            data=None,
        )
        return ListableResource(
            response=response.json(),
            cls=DataFileUploadResponse,
            auth=self.auth,
            path="/data-files",
            max_items=max_items,
            query_params=query_params,
        )

    def create(
        self,
        File: io.BufferedReader,
        name: str = None,
        appId: str = None,
        connectionId: str = None,
        sourceId: str = None,
        tempContentFileId: str = None,
    ) -> DataFileUploadResponse:
        """
        Upload a new data file.


        Parameters
        ----------
        """

        files_dict = {}
        files_dict["File"] = File

        Json_dict = {}
        if name is not None:
            Json_dict["name"] = name
        if appId is not None:
            Json_dict["appId"] = appId
        if connectionId is not None:
            Json_dict["connectionId"] = connectionId
        if sourceId is not None:
            Json_dict["sourceId"] = sourceId
        if tempContentFileId is not None:
            Json_dict["tempContentFileId"] = tempContentFileId
        files_dict["Json"] = (None, json.dumps(Json_dict))

        response = self.auth.rest(
            path="/data-files", method="POST", params={}, data=None, files=files_dict
        )
        obj = DataFileUploadResponse(**response.json())
        obj.auth = self.auth
        return obj

    def change_space(self, data: DataFileBatchChangeSpaceRequest) -> None:
        """
        Change the spaces that a set of existing data files reside in a a single batch.
        This is to allow for a separate admin type of operation that is more global in terms of access in cases
        where admin users may not explicitly have been granted full access to a given space within the declared
        space-level permissions. If the space ID is set to null, then the data file will end up residing in the
        personal space of the user who is the owner of the file.


        Parameters
        ----------
        data: DataFileBatchChangeSpaceRequest
        """

        try:
            data = asdict(data)
        except:
            data = data

        self.auth.rest(
            path="/data-files/actions/change-space",
            method="POST",
            params={},
            data=data,
        )

    def delete(self, data: DataFileBatchDeleteRequest) -> None:
        """
        Delete the specified set of data files as a single batch.


        Parameters
        ----------
        data: DataFileBatchDeleteRequest
        """

        try:
            data = asdict(data)
        except:
            data = data

        self.auth.rest(
            path="/data-files/actions/delete",
            method="POST",
            params={},
            data=data,
        )

    def get_connections(
        self,
        appId: str = None,
        name: str = None,
        spaceId: str = None,
        personal: bool = None,
        page: str = None,
        limit: int = 20,
        sort: str = None,
        max_items: int = 20,
    ) -> ListableResource[ConnectionsResponse]:
        """
        Get the list of built-in connections used by the engine to load/write data files.
        The non-filtered list contains a set of hardcoded connections, along with one connection per team space that
        the given user has access to.


        appId: str
          If present, get connections with connection strings that are scoped to the given app ID.

        name: str
          If present, only return connections with the given name.

        spaceId: str
          If present, only return the connection that accesses data files in the specified space.

        personal: bool
          If true, only return the connections that access data in a personal space. Default is false.

        page: str
          If present, the cursor that starts the page of data that is returned.

        limit: int
          If present, the maximum number of data file connection records to return.

        sort: str
          The name of the field used to sort the result. By default, the sort is ascending. Putting a '+' prefix on
          the sort field name explicitly indicates ascending sort order. A '-' prefix indicates a descending sort order.

        Parameters
        ----------
        appId: str = None
        name: str = None
        spaceId: str = None
        personal: bool = None
        page: str = None
        limit: int = 20
        sort: str = None
        """
        query_params = {}
        if appId is not None:
            query_params["appId"] = appId
        if name is not None:
            query_params["name"] = name
        if spaceId is not None:
            query_params["spaceId"] = spaceId
        if personal is not None:
            query_params["personal"] = personal
        if page is not None:
            query_params["page"] = page
        if limit is not None:
            query_params["limit"] = limit
        if sort is not None:
            query_params["sort"] = sort

        response = self.auth.rest(
            path="/data-files/connections",
            method="GET",
            params=query_params,
            data=None,
        )
        return ListableResource(
            response=response.json(),
            cls=ConnectionsResponse,
            auth=self.auth,
            path="/data-files/connections",
            max_items=max_items,
            query_params=query_params,
        )

    def get_quota(self) -> QuotaResponse:
        """
        Get quota information for the calling user.


        Parameters
        ----------
        """

        response = self.auth.rest(
            path="/data-files/quotas",
            method="GET",
            params={},
            data=None,
        )
        obj = QuotaResponse(**response.json())
        obj.auth = self.auth
        return obj

    def get(self, id: str) -> DataFileUploadResponse:
        """
        Get descriptive info for the specified data file.


        id: str
          The ID of the data file.

        Parameters
        ----------
        id: str
        """

        response = self.auth.rest(
            path="/data-files/{id}".replace("{id}", id),
            method="GET",
            params={},
            data=None,
        )
        obj = DataFileUploadResponse(**response.json())
        obj.auth = self.auth
        return obj
